/*
    This software may only be used by you under license from AT&T Corp.
    ("AT&T").  A copy of AT&T's Source Code Agreement is available at
    AT&T's Internet website having the URL:
    <http://www.research.att.com/sw/tools/graphviz/license/source.html>
    If you received this software without first entering into a license
    with AT&T, you have an infringing copy of this software and cannot use
    it without violating AT&T's intellectual property rights.
*/

#include "pathplan.h"

typedef unsigned char boolean;

typedef int (*qsort_cmpf)(const void*, const void*);
typedef int (*bsearch_cmpf) (const void *, const void *);

typedef struct Agraph_t graph_t;
typedef struct Agnode_t node_t;
typedef struct Agedge_t edge_t;
typedef struct Agsym_t attrsym_t;

typedef struct pointf {
	double		x,y;
} pointf;

typedef struct point {
	int			x,y;
} point;

typedef struct box {
	point		LL,UR;
} box;

typedef struct port_t {			/* internal edge endpoint specification */
	point		p;					/* aiming point */
	float		theta;				/* slope in radians */
	boolean	constrained,defined;
	unsigned char order;			/* for mincross */
} port_t;

typedef struct path {			/* internal specification for an edge spline */
	port_t		start,end;
	point		*ulpp, *urpp, *llpp, *lrpp;	/* tangents of near splines */
	int			nbox;				/* number of subdivisions */
	box			*boxes;				/* rectangular regions of subdivision */
	void		*data;
} path;

typedef struct bezier {
	point		*list;
	int			size;
	int			sflag, eflag;
	point		sp, ep;
} bezier;

typedef struct splines {
	bezier		*list;
	int			size;
} splines;

typedef struct polygon_t {		/* mutable shape information for a node */
	int			regular; 			/* true for symmetric shapes */
	int			peripheries;		/* number of periphery lines */
	int			sides;				/* number of sides */
	float		orientation;		/* orientation of shape (+ve degrees) */
	float		distortion;			/* distortion factor - as in trapezium */
	float		skew;				/* skew factor - as in parallelogram */
	int			option;				/* ROUNDED, DIAGONAL corners, etc. */
	pointf		*vertices;		    /* array of vertex points */
} polygon_t;

typedef struct shape_desc {		/* read-only shape descriptor */
	char		*name;							 /* as read from graph file */
	void		(*initfn)(node_t *);			/* initializes shape from node label */
	port_t		(*portfn)(node_t *, char *);	/* finds aiming point and slope of port */
	int			(*insidefn)(node_t *, pointf, edge_t*);		/* clips incident edge spline */
	int			(*pboxfn)(node_t *, edge_t *, int, box *, int *);		/* finds box path to reach port */
	void		(*codefn)(node_t *);			/* emits graphics code for node */
	polygon_t	*polygon;						/* base polygon info */
} shape_desc;

typedef struct codegen_t {
	void     (*reset)(void);
	void	(*begin_job)(FILE *ofp,graph_t *g, char **lib, char *user, char *info[], point pages);
	void 	(*end_job)(void);
	void	(*begin_graph)(graph_t* g, box bb, point pb);
	void	(*end_graph)(void);
	void	(*begin_page)(point page, double scale, int rot, point offset);
	void	(*end_page)(void);
	void    (*begin_cluster)(graph_t *),(*end_cluster)(void); 
	void    (*begin_nodes)(void), (*end_nodes)(void); 
	void    (*begin_edges)(void),(*end_edges)(void); 
	void	(*begin_node)(node_t *),(*end_node)(void);
	void	(*begin_edge)(edge_t *),(*end_edge)(void);
	void	(*begin_context)(void),(*end_context)(void);
	void	(*set_font)(char *fontname, double fontsize);
	void	(*textline)(point p, char *str, int width, double fontsz, double align);
	void	(*set_color)(char *name),(*set_style)(char **s);
	void	(*ellipse)(point p, int rx, int ry, int filled);
	void	(*polygon)(point *A, int n, int filled);
	void	(*beziercurve)(point *A, int n, int arrow_at_start, int arrow_at_end);
	void	(*polyline)(point *A,int n);
	/* void	(*arrowhead)(point p, double theta, double scale, int flag); */
	boolean bezier_has_arrows;
	void	(*user_shape)(char *name, point *A, int sides, int filled);
	void	(*comment)(void* obj, attrsym_t* sym);
	point	(*textsize)(char *str, char *fontname, double fontsz);
} codegen_t;

typedef struct queue {
	node_t	**store,**limit,**head,**tail;
} queue;

typedef struct adjmatrix_t {
	int		nrows,ncols;
	char	*data;
} adjmatrix_t;

typedef struct rank_t {
	int				n;			/* number of nodes in this rank			*/
	node_t			**v;		/* ordered list of nodes in rank		*/
	int				an;			/* globally allocated number of nodes	*/
	node_t			**av;		/* allocated list of nodes in rank		*/
	int 			ht2;		/* height above the centerline			*/
	boolean			candidate;	/* for transpose ()						*/
	boolean			valid;
	int				cache_nc;	/* caches number of crossings			*/
	adjmatrix_t		*flat;
} rank_t;

typedef struct textline_t {
	char			*str;
	short			width;
	char			just;
} textline_t;

typedef struct textlabel_t {
	char			*text,*fontname,*fontcolor;
	float			fontsize;
	pointf			dimen;
	point			p;
	textline_t		*line;
	char			nlines;
} textlabel_t;

typedef struct layout_t {
	float			quantum,scale,font_scale_adj;
	point			margin, page, size;
	boolean			landscape,centered;
} layout_t;

/* for "record" shapes */
typedef struct field_t {
	point		size;			/* its dimension */
	box			b;				/* its final placement */
	int			n_flds;	
	textlabel_t	*lp;			/* n_flds == 0 */
	struct field_t	**fld;		/* n_flds > 0 */
	int		LR;					/* if box list is horizontal (left to right) */
	char	*id;				/* user's identifier */
} field_t;

typedef struct hsbcolor_t {
	char			*name;
	unsigned char	h,s,b;
} hsbcolor_t;

typedef struct Agraphinfo_t {
		/* to generate code */
	layout_t			*drawing;
    textlabel_t        	*label;         /* if the cluster has a title */
	box					bb;				/* bounding box */

		/* not used at present, safe if set to zero */
	int					n_cluster;
	graph_t				**clust;
	boolean				left_to_right,has_edge_labels;

		/* to place nodes */
	node_t				**nlist;
	int					move;
	double				**dist,**spring,**sum_t,***t;

} Agraphinfo_t;

typedef struct Agnodeinfo_t {
	shape_desc			*shape;
	void				*shape_info;
	point				coord;
	float				width,height;
	int					ht,lw,rw;
	textlabel_t			*label;
	char				state;

	boolean				pinned;
	short				xsize,ysize;
	int					id,heapindex,hops;
	double				pos[NDIM],dist;
} Agnodeinfo_t;

typedef struct Agedgeinfo_t {
	splines				*spl;
	port_t				tail_port,head_port;	/* might be used someday */
	textlabel_t			*label,*head_label,*tail_label;

	float				factor;
	float				dist;
	edge_t				*to_orig;				/* for dot's shapes.c    */
	char				edge_type;

	Ppolyline_t			path;
} Agedgeinfo_t;
