/* 
 * gdCmd.c - Interface gd to Tcl.
 * 
 * Author:		Spencer W. Thomas
 * 				Information Technology and Networking
 * 				University of Michigan
 * Date:		Fri Aug 19 1994
 * Copyright (c) 1994, University of Michigan
 *
 * Updates by John Ellson (ellson@lucent.com):
 *	   - conversion to freetype text rendering
 *     - support for tcl8.0 Tcl_Objects
 *     - support for tcl8.1 Unicode
 */

#include <stdio.h>
#include <string.h>
#include <math.h>
#include <tcl.h>
#include "gd.h"
#include "gdhandle.h"
#include "gdttf.h"

void *GDHandleTable;

/* global data */
typedef struct {
	tblHeader_pt	 	handleTbl;
} GdData;

static int tclGdCreateCmd(), tclGdDestroyCmd(), tclGdWriteCmd(),
	tclGdColorCmd(), tclGdInterlaceCmd(), tclGdSetCmd(), tclGdLineCmd(),
	tclGdRectCmd(), tclGdArcCmd(), tclGdFillCmd(), tclGdSizeCmd(),
	tclGdTextCmd(), tclGdCopyCmd(), tclGdGetCmd(),
	tclGdBrushCmd(), tclGdStyleCmd(), tclGdTileCmd(), tclGdPolygonCmd();
static int tclGdColorNewCmd(), tclGdColorExactCmd(),
	tclGdColorClosestCmd(), tclGdColorResolveCmd(), tclGdColorFreeCmd(),
	tclGdColorTranspCmd(), tclGdColorGetCmd();

typedef  struct {
	char *cmd;
	int (*f)();
	int minargs, maxargs;
	int subcmds;
	int ishandle;
	char *usage;
} cmdOptions;

static cmdOptions subcmdVec[] = {
	{"create",			tclGdCreateCmd,			2, 2,	0, 0,
			"width height"},
	{"createFromGIF",	tclGdCreateCmd, 		1, 1,	0, 0,
			"filehandle"},
	{"createFromGD",	tclGdCreateCmd, 		1, 1,	0, 0,
			"filehandle"},
	{"createFromXBM",	tclGdCreateCmd, 		1, 1,	0, 0,
			"filehandle"},

	{"destroy",			tclGdDestroyCmd,		1, 1,	0, 1,
			"gdhandle"},

	{"writeGIF",		tclGdWriteCmd,			2, 2,	0, 1,
			"gdhandle filehandle"},
	{"writeGD",			tclGdWriteCmd,			2, 2,	0, 1,
			"gdhandle filehandle"},

	{"interlace",		tclGdInterlaceCmd,		1, 2,	0, 1,
		 "gdhandle ?on-off?"},

	{"color",			tclGdColorCmd,			2, 5,	1, 1,
			"option values..."},
	{"brush",			tclGdBrushCmd,			2, 2,	0, 2,
			"gdhandle brushhandle"},
	{"style",			tclGdStyleCmd,			2, 999,	0, 1,
			"gdhandle color..."},
	{"tile",			tclGdTileCmd,			2, 2,	0, 2,
			"gdhandle tilehandle"},

	{"set",				tclGdSetCmd,			4, 4,	0, 1,
		"gdhandle color x y"},
	{"line",			tclGdLineCmd,			6, 6,	0, 1,
			"gdhandle color x1 y1 x2 y2"},
	{"rectangle",		tclGdRectCmd,			6, 6,	0, 1,
			"gdhandle color x1 y1 x2 y2"},
	{"fillrectangle",	tclGdRectCmd,			6, 6,	0, 1,
			"gdhandle color x1 y1 x2 y2"},
	{"arc",				tclGdArcCmd,			8, 8, 	0, 1,
			"gdhandle color cx cy width height start end"},
	{"fillarc",			tclGdArcCmd,			8, 8, 	0, 1,
			"gdhandle color cx cy width height start end"},
	{"polygon",			tclGdPolygonCmd,		2, 999,	0, 1,
			"gdhandle color x1 y1 x2 y2 x3 y3 ..."},
	{"fillpolygon",		tclGdPolygonCmd,		3, 999,	0, 1,
			"gdhandle color x1 y1 x2 y2 x3 y3 ..."},
	{"fill",			tclGdFillCmd,			4, 5,	0, 1,
			"gdhandle color x y ?bordercolor?"},
/* 
 * we allow null gd handles to the text command to allow program to get size
 * of text string, so the text command provides its own handle processing and checking
 */
	{"text",			tclGdTextCmd,			8, 8,	0, 0,
			"gdhandle color fontpathname size angle x y string"},


	{"copy",			tclGdCopyCmd,			8, 10,	0, 2,
			"desthandle srchandle destx desty srcx srcy destw desth ?srcw srch?"},

	{"get",				tclGdGetCmd,			3, 3,	0, 1,
			"gdhandle x y"},
	{"size",			tclGdSizeCmd,			1, 1,	0, 1,
			"gdhandle"},
};

static cmdOptions colorCmdVec[] = {
	{"new",				tclGdColorNewCmd,		5, 5,	1, 1,
			"gdhandle red green blue"},
	{"exact",			tclGdColorExactCmd,		5, 5,	1, 1,
			"gdhandle red green blue"},
	{"closest",			tclGdColorClosestCmd,	5, 5,	1, 1,
			"gdhandle red green blue"},
	{"resolve",			tclGdColorResolveCmd,	5, 5,	1, 1,
			"gdhandle red green blue"},
	{"free",			tclGdColorFreeCmd,		3, 3,	1, 1,
			"gdhandle color"},
	{"transparent",		tclGdColorTranspCmd,	2, 3,	1, 1,
			"gdhandle ?color?"},
	{"get",				tclGdColorGetCmd,		2, 3,	1, 1,
			"gdhandle ?color?"}
};

#if (TCL_MAJOR_VERSION > 7)
#define TCLOBJ
#if !(defined(Tcl_GetString))
/* Tcl_GetString didn't appear until tcl8.1 */
char *
Tcl_GetString(Tcl_Obj *obj) {
	int len;
	if (!obj) {
		return NULL;
	}
	return (Tcl_GetStringFromObj(obj, &len));
}
#endif
#endif

/*
 * Helper function to interpret color_idx values.
 */
static int
#ifndef TCLOBJ
tclGd_GetColor(Tcl_Interp *interp, char *arg, int *color)
#else /* TCLOBJ */
tclGd_GetColor(Tcl_Interp *interp, Tcl_Obj *obj, int *color)
#endif /* TCLOBJ */
{
	int nlist, retval = TCL_OK;
#ifndef TCLOBJ
	char **theList;
#else /* TCLOBJ */
	Tcl_Obj **theList;
	char *firsttag, *secondtag;
#endif /* TCLOBJ */

	/* Assume it's an integer, check other cases on failure. */
#ifndef TCLOBJ
	if (Tcl_GetInt(interp, arg, color) == TCL_OK)
#else /* TCLOBJ */
	if (Tcl_GetIntFromObj(interp, obj, color) == TCL_OK)
#endif /* TCLOBJ */
		return TCL_OK;
	else
	{
		Tcl_ResetResult(interp);
#ifndef TCLOBJ
		Tcl_SplitList(interp, arg, &nlist, &theList);
#else /* TCLOBJ */
		if (Tcl_ListObjGetElements(interp, obj, &nlist, &theList) != TCL_OK)
			return TCL_ERROR;
#endif /* TCLOBJ */
		if (nlist < 1 || nlist > 2)
			retval = TCL_ERROR;
		else
		{
#ifndef TCLOBJ
			switch (theList[0][0]) {
#else /* TCLOBJ */
			firsttag = Tcl_GetString(theList[0]);
			switch (firsttag[0]) {
#endif /* TCLOBJ */
			case 'b':
				*color = gdBrushed;
				if (nlist == 2) {
#ifndef TCLOBJ
					if (theList[1][0] == 's') {
#else /* TCLOBJ */
					secondtag = Tcl_GetString(theList[1]);
					if (secondtag[0] == 's') {
#endif /* TCLOBJ */
						*color = gdStyledBrushed;
					} else {
						retval = TCL_ERROR;
					}
				}
				break;

			case 's':
				*color = gdStyled;
				if (nlist == 2) {
#ifndef TCLOBJ
					if (theList[1][0] == 'b') {
#else /* TCLOBJ */
					secondtag = Tcl_GetString(theList[1]);
					if (secondtag[0] == 'b') {
#endif /* TCLOBJ */
						*color = gdStyledBrushed;
					} else {
						retval = TCL_ERROR;
					}
				}
				break;

			case 't':
				*color = gdTiled;
				break;

			default:
				retval = TCL_ERROR;
			}
		}
	}
	if (retval == TCL_ERROR)
#ifndef TCLOBJ
		sprintf(interp->result, "Malformed special color value");
#else /* TCLOBJ */
		Tcl_SetResult(interp, "Malformed special color value", TCL_STATIC);
#endif /* TCLOBJ */

	return retval;
}

/*
 * GD composite command:
 *
 * gd create <width> <height>
 * 		Return a handle to a new gdImage that is width X height.
 * gd createFromGIF <filehandle>
 * gd createFromGD <filehandle>
 * gd createFromXBM <filehandle>
 * 		Return a handle to a new gdImage created by reading a GIF
 * 		(resp. GD or XBM) image from the file open on filehandle.
 *
 * gd destroy <gdhandle>
 * 		Destroy the gdImage referred to by gdhandle.
 *
 * gd writeGIF <gdhandle> <filehandle>
 * gd writeGD  <gdhandle> <filehandle>
 * 		Write the image in gdhandle to filehandle as a GIF (resp. GD)
 * 		file.
 *
 * gd color new <gdhandle> <red> <green> <blue>
 * 		Allocate a new color with the given RGB values.  Returns the
 * 		color_idx, or -1 on failure (256 colors already allocated).
 * gd color exact <gdhandle> <red> <green> <blue>
 * 		Find a color_idx in the image that exactly matches the given RGB color.
 *      Returns the color_idx, or -1 if no exact match.
 * gd color closest <gdhandle> <red> <green> <blue>
 * 		Find a color in the image that is closest to the given RGB color.  
 *      Guaranteed to return a color idx.
 * gd color resolve <gdhandle> <red> <green> <blue>
 * 		Return the index of the best possible effort to get a color.  Guaranteed
 *      to return a color idx.   Equivalent to:
 *          if {[set idx [gd color exact $gd $r $g $b]] == -1} {
 *              if {[set idx [gd color neW $Gd $r $g $b]] == -1} {
 *                  set idx [gd color closest $gd $r $g $b]
 *              }
 *          }
 * gd color free <gdhandle> <color_idx>
 * 		Free the color at the given color_idx for reuse.
 * gd color transparent <gdhandle> <color_idx>
 * 		Mark the color_idx as the transparent background color.
 * gd color get <gdhandle> [<color_idx>]
 * 		Return the RGB value at <color_idx>, or {} if it is not allocated.
 * 		If <color_idx> is not specified, return a list of {color_idx R G B} values
 * 		for all allocated colors.
 * gd color gettransparent <gdhandle>
 * 		Return the color_idx of the transparent color.
 * 
 * gd brush <gdhandle> <brushhandle>
 * 		Set the brush image to be used for brushed lines.  Transparent
 * 		pixels in the brush will not change the image when the brush
 * 		is applied. 
 * gd style <gdhandle> <color_idx> ...
 * 		Set the line style to the list of color indices.  This is interpreted
 * 		in one of two ways.  For a simple styled line, each color is
 * 		applied to points along the line in turn.  The transparent color
 * 		value may be used to leave gaps in the line.  For a styled, brushed
 * 		line, a 0 (or the transparent color_idx) means not to fill the pixel,
 * 		and a non-zero value means to apply the brush.
 * gd tile <gdhandle> <tilehandle>
 * 		Set the tile image to be used for tiled fills.  Transparent pixels in
 * 		the tile will not change the underlying image during tiling.
 * 
 * In all drawing functions, the color_idx is a number, or may be one of the
 * strings styled, brushed, tiled, "styled brushed" or "brushed styled".  The
 * style, brush, or tile currently in effect will be used.  Brushing and
 * styling apply to lines, tiling to filled areas.
 * 
 * gd set <gdhandle> <color_idx> <x> <y>
 * 		Set the pixel at (x,y) to color <color_idx>.
 * gd line <gdhandle> <color_idx> <x1> <y1> <x2> <y2>
 * 		Draw a line in color <color_idx> from (x1,y1) to (x2,y2).
 * gd rectangle <gdhandle> <color_idx> <x1> <y1> <x2> <y2>
 * gd fillrectangle <gdhandle> <color_idx> <x1> <y1> <x2> <y2>
 * 		Draw the outline of (resp. fill) a rectangle in color <color_idx>
 * 		with corners at (x1,y1) and (x2,y2).
 * gd arc <gdhandle> <color_idx> <cx> <cy> <width> <height> <start> <end>
 * gd fillarc <gdhandle> <color_idx> <cx> <cy> <width> <height> <start> <end>
 * 		Draw an arc, or filled segment, in color <color_idx>, centered at (cx,cy)
 *      in a rectangle width x height, starting at start degrees and ending 
 *      at end degrees.		Start must be > end.
 * gd polygon <gdhandle> <color_idx> <x1> <y1> ...
 * gd fillpolygon <gdhandle> <color_idx> <x1> <y1> ...
 * 		Draw the outline of, or fill, a polygon specified by the x, y
 * 		coordinate list.  
 * 
 * gd fill <gdhandle> <color_idx> <x> <y>
 * gd fill <gdhandle> <color_idx> <x> <y> <borderindex>
 * 		Fill with color <color_idx>, starting from (x,y) within a region
 * 		of pixels all the color of the pixel at (x,y) (resp., within a
 * 		border colored borderindex).
 *
 * gd size <gdhandle>
 * 		Returns a list {width height} of the image.
 *
 * gd text <gdhandle> <color_idx> <fontpath> <size> <angle> <x> <y> <string>
 *      Draw text using the .ttf font in <fontpath> in color <color_idx>,
 *      with pointsize <size>, rotation in radians <angle>, with lower left 
 *      corner at (x,y).  String may contain UTF8 sequences like: "&#192;"
 *      Returns 4 corner coords of bounding rectangle.
 *      Use gdhandle = {} to get boundary without rendering.
 *      Use negative of color_idx to disable antialiasing.
 *
 * gd copy <desthandle> <srchandle> <destx> <desty> <srcx> <srcy> <w> <h>
 * gd copy <desthandle> <srchandle> <destx> <desty> <srcx> <srcy> \
 * 				<destw> <desth> <srcw> <srch>
 * 		Copy a subimage from srchandle(srcx, srcy) to
 * 		desthandle(destx, desty), size w x h.  Or, resize the subimage
 * 		in copying from srcw x srch to destw x desth.
 * 
 */
#ifdef TCLOBJ

#endif /* TCLOBJ */
int
#ifndef TCLOBJ
gdCmd(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
#else /* TCLOBJ */
gdCmd(ClientData clientData, Tcl_Interp *interp, 
	int argc, Tcl_Obj *CONST objv[])
#endif /* TCLOBJ */
{
	GdData *gdData = (GdData *)clientData;
	int argi, subi;
#ifdef TCLOBJ
	char buf[100];
#endif /* TCLOBJ */
	/* Check for subcommand. */
	if (argc < 2) {
#ifndef TCLOBJ
		interp->result = "wrong # args: should be \"gd option ...\"";
#else /* TCLOBJ */
		Tcl_SetResult(interp,
			"wrong # args: should be \"gd option ...\"", TCL_STATIC);
#endif /* TCLOBJ */
		return TCL_ERROR;
	}
	
	/* Find the subcommand. */
	for (subi = 0; subi < (sizeof subcmdVec) / (sizeof subcmdVec[0]); subi++) {
#ifndef TCLOBJ
		if (strcmp(subcmdVec[subi].cmd, argv[1]) == 0) {
#else /* TCLOBJ */
		if (strcmp(subcmdVec[subi].cmd, Tcl_GetString(objv[1])) == 0) {
#endif /* TCLOBJ */

			/* Check arg count. */
			if (argc - 2 < subcmdVec[subi].minargs ||
				 argc - 2 > subcmdVec[subi].maxargs) {
#ifndef TCLOBJ
				Tcl_AppendResult(interp, "wrong # args: should be \"gd ",
								  subcmdVec[subi].cmd, " ",
								  subcmdVec[subi].usage, "\"", 0);
#else /* TCLOBJ */
				sprintf(buf, "wrong # args: should be \"gd %s %s\"",
					subcmdVec[subi].cmd,  subcmdVec[subi].usage);
				Tcl_SetResult(interp, buf, TCL_VOLATILE);
#endif /* TCLOBJ */
				return TCL_ERROR;
			}

			/* Check for valid handle(s). */
			if (subcmdVec[subi].ishandle > 0) {
				/* Are any handles allocated? */
				if (gdData->handleTbl == NULL) {
#ifndef TCLOBJ
					sprintf(interp->result, "no such handle%s: ",
#else /* TCLOBJ */
					sprintf(buf, "no such handle%s: ",
#endif /* TCLOBJ */
							 subcmdVec[subi].ishandle > 1 ? "s" : "");
#ifdef TCLOBJ
					Tcl_SetResult(interp, buf, TCL_VOLATILE);
#endif /* TCLOBJ */
					for (argi = 2 + subcmdVec[subi].subcmds;
						  argi < 2 + subcmdVec[subi].subcmds +
						  		subcmdVec[subi].ishandle;
						  argi++) {
#ifndef TCLOBJ
						Tcl_AppendResult(interp, argv[argi], " ", 0);
#else /* TCLOBJ */
						Tcl_AppendResult(interp, 
							Tcl_GetString(objv[argi]), " ", 0);
#endif /* TCLOBJ */
					}
					return TCL_ERROR;
				}
				/* Check each handle to see if it's a valid handle. */
				if (2+subcmdVec[subi].subcmds+subcmdVec[subi].ishandle > argc) {
#ifndef TCLOBJ
					interp->result = "GD handle(s) not specified";
#else /* TCLOBJ */
					Tcl_SetResult(interp, "GD handle(s) not specified", TCL_STATIC);
#endif /* TCLOBJ */
					return TCL_ERROR;
				}
				for (argi = 2 + subcmdVec[subi].subcmds;
					  argi < (2 + subcmdVec[subi].subcmds +
							  subcmdVec[subi].ishandle);
					  argi++) {
#ifndef TCLOBJ
					if (! gdHandleXlate(interp, gdData->handleTbl, argv[argi]))
#else /* TCLOBJ */
					if (! gdHandleXlate(interp, gdData->handleTbl, 
						Tcl_GetString(objv[argi])))
#endif /* TCLOBJ */
						return TCL_ERROR;
				}
			}
			
			/* Call the subcommand function. */
#ifndef TCLOBJ
			return (*subcmdVec[subi].f)(interp, gdData, argc, argv);
#else /* TCLOBJ */
			return (*subcmdVec[subi].f)(interp, gdData, argc, objv);
#endif /* TCLOBJ */
		}
	}

	/* If we get here, the option doesn't match. */
#ifndef TCLOBJ
	Tcl_AppendResult(interp, "bad option \"", argv[1], "\": should be ", 0);
#else /* TCLOBJ */
	Tcl_AppendResult(interp, "bad option \"", 
		Tcl_GetString(objv[1]), "\": should be ", 0);
#endif /* TCLOBJ */
	for (subi = 0; subi < (sizeof subcmdVec) / (sizeof subcmdVec[0]); subi++)
		Tcl_AppendResult(interp, (subi > 0 ? ", " : ""),
						   subcmdVec[subi].cmd, 0);
#ifndef TCLOBJ

#endif /* not TCLOBJ */
	return TCL_ERROR;
}

static int
tclGdCreateCmd(Tcl_Interp *interp, GdData *gdData, 
#ifndef TCLOBJ
	int argc, char *argv[])
#else /* TCLOBJ */
	int argc, Tcl_Obj *CONST objv[])
#endif /* TCLOBJ */
{
	int w, h;
	gdImagePtr im;
	ClientData filePtr;
#ifdef TCLOBJ
	char *cmd, buf[50];
#endif /* TCLOBJ */

#ifndef TCLOBJ
	if (strcmp(argv[1], "create") == 0) {
		if (Tcl_GetInt(interp, argv[2], &w) != TCL_OK)
#else /* TCLOBJ */
	cmd = Tcl_GetString(objv[1]);
	if (strcmp(cmd, "create") == 0) {
		if (Tcl_GetIntFromObj(interp, objv[2], &w) != TCL_OK)
#endif /* TCLOBJ */
			return TCL_ERROR;
#ifndef TCLOBJ
		if (Tcl_GetInt(interp, argv[3], &h) != TCL_OK)
#else /* TCLOBJ */
		if (Tcl_GetIntFromObj(interp, objv[3], &h) != TCL_OK)
#endif /* TCLOBJ */
			return TCL_ERROR;
		im = gdImageCreate(w, h);
		if (im == NULL)
		{
#ifndef TCLOBJ
			sprintf(interp->result, "GD unable to allocate %d X %d image",
					 w, h);
#else /* TCLOBJ */
			sprintf(buf, "GD unable to allocate %d X %d image", w, h);
			Tcl_SetResult(interp, buf, TCL_VOLATILE);
#endif /* TCLOBJ */
			return TCL_ERROR;
		}
	} else {
#ifndef TCLOBJ
		if (Tcl_GetOpenFile(interp, argv[2], 0, 1, &filePtr) != TCL_OK)
#else /* TCLOBJ */
		if (Tcl_GetOpenFile(interp, Tcl_GetString(objv[2]),
				0, 1, &filePtr) != TCL_OK)
#endif /* TCLOBJ */
			return TCL_ERROR;
		/* Read GIF, XBM, or GD file? */
#ifndef TCLOBJ
		if (argv[1][11] == 'I') {
#else /* TCLOBJ */
		if (cmd[11] == 'I') {
#endif /* TCLOBJ */
			im = gdImageCreateFromGif((FILE *)filePtr);
#ifndef TCLOBJ
		} else if (argv[1][10] == 'X') {
#else /* TCLOBJ */
		} else if (cmd[10] == 'X') {
#endif /* TCLOBJ */
			im = gdImageCreateFromXbm((FILE *)filePtr);
		} else {
			im = gdImageCreateFromGd((FILE *)filePtr);
		}
		if (im == NULL) {
#ifndef TCLOBJ
			interp->result = "GD unable to read image file";
#else /* TCLOBJ */
			Tcl_SetResult(interp,"GD unable to read image file", TCL_STATIC);
#endif /* TCLOBJ */
			return TCL_ERROR;
		}
	}

#ifndef TCLOBJ
	*(gdImagePtr *)(gdHandleAlloc(gdData->handleTbl, interp->result)) = im;
#else /* TCLOBJ */
	*(gdImagePtr *)(gdHandleAlloc(gdData->handleTbl, buf)) = im;
	Tcl_SetResult(interp, buf, TCL_VOLATILE);
#endif /* TCLOBJ */
	return TCL_OK;
}

static int
#ifndef TCLOBJ
tclGdDestroyCmd(Tcl_Interp *interp, GdData *gdData, int argc, char *argv[])
#else /* TCLOBJ */
tclGdDestroyCmd(Tcl_Interp *interp, GdData *gdData, 
	int argc, Tcl_Obj *CONST objv[])
#endif /* TCLOBJ */
{
	gdImagePtr im;
	void *hdl;

	/* Get the handle, and the image pointer. */
#ifndef TCLOBJ
	hdl = (void *)gdHandleXlate(interp, gdData->handleTbl, argv[2]);
#else /* TCLOBJ */
	hdl = (void *)gdHandleXlate(interp, gdData->handleTbl,
		Tcl_GetString(objv[2]));
#endif /* TCLOBJ */
	im = *(gdImagePtr *)hdl;
	/* Release the handle, destroy the image. */
	gdHandleFree(gdData->handleTbl, hdl);
	gdImageDestroy(im);

	return TCL_OK;
}

static int
#ifndef TCLOBJ
tclGdWriteCmd(Tcl_Interp *interp, GdData *gdData, int argc, char *argv[])
#else /* TCLOBJ */
tclGdWriteCmd(Tcl_Interp *interp, GdData *gdData, 
	int argc, Tcl_Obj *CONST objv[])
#endif /* TCLOBJ */
{
	gdImagePtr im;
	ClientData filePtr;
#ifndef TCLOBJ
	
#else /* TCLOBJ */
	char *cmd;

	cmd = Tcl_GetString(objv[1]);
#endif /* TCLOBJ */
	/* Get the image pointer. */
#ifndef TCLOBJ
	im = *(gdImagePtr *)gdHandleXlate(interp, gdData->handleTbl, argv[2]);
#else /* TCLOBJ */
	im = *(gdImagePtr *)gdHandleXlate(interp, gdData->handleTbl, 
		Tcl_GetString(objv[2]));
#endif /* TCLOBJ */

	/* Get the file reference. */
#ifndef TCLOBJ
	if (Tcl_GetOpenFile(interp, argv[3], 1, 1, &filePtr) != TCL_OK)
#else /* TCLOBJ */
	if (Tcl_GetOpenFile(interp, Tcl_GetString(objv[3]), 
			1, 1, &filePtr) != TCL_OK)
#endif /* TCLOBJ */
		return TCL_ERROR;

	/* Do it. */
#ifndef TCLOBJ
	if (argv[1][6] == 'I') {
#else /* TCLOBJ */
	if (cmd[6] == 'I') {
#endif /* TCLOBJ */
		gdImageGif(im, (FILE *)filePtr);
	} else {
		gdImageGd(im, (FILE *)filePtr);
	}
	fflush((FILE *)filePtr);
	return TCL_OK;
}

static int
#ifndef TCLOBJ
tclGdInterlaceCmd(Tcl_Interp *interp, GdData *gdData, int argc, char *argv[])
#else /* TCLOBJ */
tclGdInterlaceCmd(Tcl_Interp *interp, GdData *gdData, 
	int argc, Tcl_Obj *CONST objv[])
#endif /* TCLOBJ */
{
	gdImagePtr im;
	int on_off;

	/* Get the image pointer. */
#ifndef TCLOBJ
	im = *(gdImagePtr *)gdHandleXlate(interp, gdData->handleTbl, argv[2]);
#else /* TCLOBJ */
	im = *(gdImagePtr *)gdHandleXlate(interp, gdData->handleTbl, 
		Tcl_GetString(objv[2]));
#endif /* TCLOBJ */

	if (argc == 4) {
		/* Get the on_off values. */
#ifndef TCLOBJ
		if (Tcl_GetBoolean(interp, argv[3], &on_off) != TCL_OK)
#else /* TCLOBJ */
		if (Tcl_GetBooleanFromObj(interp, objv[3], &on_off) != TCL_OK)
#endif /* TCLOBJ */
			return TCL_ERROR;

		/* Do it. */
		gdImageInterlace(im, on_off);
	} else {
		/* Get the current state. */
#ifndef TCLOBJ
		sprintf(interp->result, "%d", gdImageGetInterlaced(im));
#else /* TCLOBJ */
		on_off = gdImageGetInterlaced(im);
#endif /* TCLOBJ */
	}
#ifndef TCLOBJ

#else /* TCLOBJ */
	Tcl_SetObjResult(interp, Tcl_NewBooleanObj(on_off));
#endif /* TCLOBJ */
	return TCL_OK;
}


static int
#ifndef TCLOBJ
tclGdColorCmd(Tcl_Interp *interp, GdData *gdData, int argc, char *argv[])
#else /* TCLOBJ */
tclGdColorCmd(Tcl_Interp *interp, GdData *gdData, 
	int argc, Tcl_Obj *CONST objv[])
#endif /* TCLOBJ */
{
	gdImagePtr im;
#ifndef TCLOBJ
	int subi, i, nsub, args[3];

#else /* TCLOBJ */
	int subi, nsub, i, args[3];
#endif /* TCLOBJ */

	nsub = (sizeof colorCmdVec) / (sizeof colorCmdVec[0]);
	if (argc >= 3) {
		/* Find the subcommand. */
		for (subi = 0; subi < nsub; subi++)
		{
#ifndef TCLOBJ
			if (strcmp(colorCmdVec[subi].cmd, argv[2]) == 0)
#else /* TCLOBJ */
			if (strcmp(colorCmdVec[subi].cmd, Tcl_GetString(objv[2])) == 0)
#endif /* TCLOBJ */
			{
				/* Check arg count. */
				if (argc - 2 < colorCmdVec[subi].minargs ||
					 argc - 2 > colorCmdVec[subi].maxargs)
				{
					Tcl_AppendResult(interp,
									  "wrong # args: should be \"gd color ",
									  colorCmdVec[subi].cmd, " ",
									  colorCmdVec[subi].usage, "\"", 0);
					return TCL_ERROR;
				}

				/* Get the image pointer. */
#ifndef TCLOBJ
				im = *(gdImagePtr *)gdHandleXlate(interp, gdData->handleTbl, argv[3]);
			
#else /* TCLOBJ */
				im = *(gdImagePtr *)gdHandleXlate(interp, gdData->handleTbl,
									 Tcl_GetString(objv[3]));
#endif /* TCLOBJ */
				/* Parse off integer arguments.
				 * 1st 4 are gd color <opt> <handle>
				 */
				for (i = 0; i < argc - 4; i++) {
#ifndef TCLOBJ
					if (Tcl_GetInt(interp, argv[i+4], &args[i]) != TCL_OK) {
#else /* TCLOBJ */
					if (Tcl_GetIntFromObj(interp, objv[i+4], &args[i]) != TCL_OK) {
#endif /* TCLOBJ */


#if 0
						if (args[i] < 0 || args[i] > 255) {
#else
/* gd text uses -ve colors to turn off anti-aliasing */
						if (args[i] < -255 || args[i] > 255) {
#endif
#ifndef TCLOBJ
							interp->result = "argument out of range 0-255";
#else /* TCLOBJ */
							Tcl_SetResult(interp, "argument out of range 0-255", TCL_STATIC);
#endif /* TCLOBJ */
							return TCL_ERROR;
						}
					}
				}

				/* Call the subcommand function. */
				return (*colorCmdVec[subi].f)(interp, im, argc - 4, args);
			}
		}
	}

	/* If we get here, the option doesn't match. */
	if (argc > 2) {
#ifndef TCLOBJ
		Tcl_AppendResult(interp, "bad option \"", argv[2], "\": ", 0);
#else /* TCLOBJ */
		Tcl_AppendResult(interp, "bad option \"", 
				Tcl_GetString(objv[2]), "\": ", 0);
#endif /* TCLOBJ */
	} else {
		Tcl_AppendResult(interp, "wrong # args: ", 0);
	}
	Tcl_AppendResult(interp, "should be ", 0);
	for (subi = 0; subi < nsub; subi++)
		Tcl_AppendResult(interp, (subi > 0 ? ", " : ""),
						   colorCmdVec[subi].cmd, 0);

	return TCL_ERROR;
}

static int
tclGdColorNewCmd(Tcl_Interp *interp, gdImagePtr im, int argc, int args[])
{
	int color;

#ifndef TCLOBJ
	color = gdImageColorAllocate(im, args[0], args[1], args[2]);

	sprintf(interp->result, "%d", color);
#else /* TCLOBJ */
	color = gdImageColorAllocate(im, args[0], args[1], args[2]); 
    Tcl_SetObjResult(interp, Tcl_NewIntObj(color));
#endif /* TCLOBJ */
	return TCL_OK;
}

static int
tclGdColorExactCmd(Tcl_Interp *interp, gdImagePtr im, int argc, int args[])
{
	int color;

#ifndef TCLOBJ
	color = gdImageColorExact(im, args[0], args[1], args[2]);

	sprintf(interp->result, "%d", color);
#else /* TCLOBJ */
	color = gdImageColorExact(im, args[0], args[1], args[2]); 
    Tcl_SetObjResult(interp, Tcl_NewIntObj(color));
#endif /* TCLOBJ */
	return TCL_OK;
}

static int
tclGdColorClosestCmd(Tcl_Interp *interp, gdImagePtr im, int argc, int args[])
{
	int color;

#ifndef TCLOBJ
	color = gdImageColorClosest(im, args[0], args[1], args[2]);

	sprintf(interp->result, "%d", color);
#else /* TCLOBJ */
	color = gdImageColorClosest(im, args[0], args[1], args[2]); 
    Tcl_SetObjResult(interp, Tcl_NewIntObj(color));
#endif /* TCLOBJ */
	return TCL_OK;
}

static int
tclGdColorResolveCmd(Tcl_Interp *interp, gdImagePtr im, int argc, int args[])
{
	int color;

#ifndef TCLOBJ
	color = gdImageColorResolve(im, args[0], args[1], args[2]);

	sprintf(interp->result, "%d", color);
#else /* TCLOBJ */
	color = gdImageColorResolve(im, args[0], args[1], args[2]); 
    Tcl_SetObjResult(interp, Tcl_NewIntObj(color));
#endif /* TCLOBJ */
	return TCL_OK;
}

static int
tclGdColorFreeCmd(Tcl_Interp *interp, gdImagePtr im, int argc, int args[])
{
	gdImageColorDeallocate(im, args[0]);
#ifndef TCLOBJ

#endif /* not TCLOBJ */
	return TCL_OK;
}

static int
tclGdColorTranspCmd(Tcl_Interp *interp, gdImagePtr im, int argc, int args[])
{
#ifndef TCLOBJ
	if (argc > 0) 
		gdImageColorTransparent(im, args[0]);
	else
		sprintf(interp->result, "%d", gdImageGetTransparent(im));
#else /* TCLOBJ */
	int color;
#endif /* TCLOBJ */

#ifdef TCLOBJ
	if (argc > 0) {
		color = args[0];
		gdImageColorTransparent(im, color);
	}
	else {
		color = gdImageGetTransparent(im);
	}
	Tcl_SetObjResult(interp, Tcl_NewIntObj(color));
#endif /* TCLOBJ */
	return TCL_OK;
}

static int
tclGdColorGetCmd(Tcl_Interp *interp, gdImagePtr im, int argc, int args[])
{
	char buf[30];
	int i;

	/* IF one arg, return the single color, else return list of all colors. */
	if (argc == 1)
	{
#ifndef TCLOBJ
		if (args[0] >= gdImageColorsTotal(im) || im->open[args[0]])
			return TCL_OK;		/* No such color */
		sprintf(interp->result, "%d %d %d",
				 gdImageRed(im,args[0]),
				 gdImageGreen(im,args[0]),
				 gdImageBlue(im,args[0]));
#else /* TCLOBJ */
		i = args[0];
		if (i >= gdImageColorsTotal(im) || im->open[i]) {
			Tcl_SetResult(interp, "No such color", TCL_STATIC);
			return TCL_ERROR;
		}
		sprintf(buf, "%d %d %d %d", i,
				 gdImageRed(im,i),
				 gdImageGreen(im,i),
				 gdImageBlue(im,i));
		Tcl_SetResult(interp, buf, TCL_VOLATILE);					
#endif /* TCLOBJ */
	} else {
		for (i = 0; i < gdImageColorsTotal(im); i++)
		{
			if (im->open[i])
				continue;
			sprintf(buf, "%d %d %d %d", i,
					 gdImageRed(im,i),
					 gdImageGreen(im,i),
					 gdImageBlue(im,i));
			Tcl_AppendElement(interp, buf);
		}
	}

	return TCL_OK;
}

static int 
#ifndef TCLOBJ
tclGdBrushCmd(Tcl_Interp *interp, GdData *gdData, int argc, char *argv[])
#else /* TCLOBJ */
tclGdBrushCmd(Tcl_Interp *interp, GdData *gdData,
    int argc, Tcl_Obj *CONST objv[])
#endif /* TCLOBJ */
{
	gdImagePtr im, imbrush;
	
	/* Get the image pointers. */
#ifndef TCLOBJ
	im = *(gdImagePtr *)gdHandleXlate(interp, gdData->handleTbl, argv[2]);
	imbrush = *(gdImagePtr *)gdHandleXlate(interp, gdData->handleTbl, argv[3]);
#else /* TCLOBJ */
	im = *(gdImagePtr *)gdHandleXlate(interp, gdData->handleTbl, 
		Tcl_GetString(objv[2]));
	imbrush = *(gdImagePtr *)gdHandleXlate(interp, gdData->handleTbl,
		Tcl_GetString(objv[3]));
#endif /* TCLOBJ */

	/* Do it. */
	gdImageSetBrush(im, imbrush);

	return TCL_OK;
}

static int 
#ifndef TCLOBJ
tclGdTileCmd(Tcl_Interp *interp, GdData *gdData, int argc, char *argv[])
#else /* TCLOBJ */
tclGdTileCmd(Tcl_Interp *interp, GdData *gdData,
    int argc, Tcl_Obj *CONST objv[])
#endif /* TCLOBJ */
{
	gdImagePtr im, tile;
	
	/* Get the image pointers. */
#ifndef TCLOBJ
	im = *(gdImagePtr *)gdHandleXlate(interp, gdData->handleTbl, argv[2]);
	tile = *(gdImagePtr *)gdHandleXlate(interp, gdData->handleTbl, argv[3]);
#else /* TCLOBJ */
	im = *(gdImagePtr *)gdHandleXlate(interp, gdData->handleTbl, 
		Tcl_GetString(objv[2]));
	tile = *(gdImagePtr *)gdHandleXlate(interp, gdData->handleTbl, 
		Tcl_GetString(objv[3]));
#endif /* TCLOBJ */

	/* Do it. */
	gdImageSetTile(im, tile);

	return TCL_OK;
}


static int
#ifndef TCLOBJ
tclGdStyleCmd(Tcl_Interp *interp, GdData *gdData, int argc, char *argv[])
#else /* TCLOBJ */
tclGdStyleCmd(Tcl_Interp *interp, GdData *gdData, 
    int argc, Tcl_Obj *CONST objv[])
#endif /* TCLOBJ */
{
	gdImagePtr im;
	int ncolor, *colors = NULL, i;
#ifndef TCLOBJ
	char **colorArgv = argv+3;		/* By default, colors are listed in argv. */
#else /* TCLOBJ */
	Tcl_Obj **colorObjv = (Tcl_Obj **)(&objv[3]); /* By default, colors are listed in objv. */
#endif /* TCLOBJ */
	int retval = TCL_OK;
	
	/* Get the image pointer. */
#ifndef TCLOBJ
	im = *(gdImagePtr *)gdHandleXlate(interp, gdData->handleTbl, argv[2]);
#else /* TCLOBJ */
	im = *(gdImagePtr *)gdHandleXlate(interp, gdData->handleTbl, 
		Tcl_GetString(objv[2])); 
#endif /* TCLOBJ */

	/* Figure out how many colors in the style list and allocate memory. */
	ncolor = argc - 3;
	/* If only one argument, treat it as a list. */
	if (ncolor == 1)
#ifndef TCLOBJ
		if (Tcl_SplitList(interp, argv[3], &ncolor, &colorArgv) != TCL_OK)
#else /* TCLOBJ */
		if (Tcl_ListObjGetElements(interp, objv[3],
				&ncolor, &colorObjv) != TCL_OK)
#endif /* TCLOBJ */
			return TCL_ERROR;

	colors = (int *)Tcl_Alloc(ncolor * sizeof(int));
	if (colors == NULL)
	{
#ifndef TCLOBJ
		sprintf(interp->result, "Memory allocation failed");
#else /* TCLOBJ */
		Tcl_SetResult(interp, "Memory allocation failed", TCL_STATIC);
#endif /* TCLOBJ */
		retval = TCL_ERROR;
		goto out;
	}
	/* Get the color values. */
	for (i = 0; i < ncolor; i++)
#ifndef TCLOBJ
		if (Tcl_GetInt(interp, colorArgv[i], &colors[i]) != TCL_OK)
#else /* TCLOBJ */
		if (Tcl_GetIntFromObj(interp, colorObjv[i], &colors[i]) != TCL_OK)
#endif /* TCLOBJ */
		{
			retval = TCL_ERROR;
			break;
		}

	/* Call the Style function if no error. */
	if (retval == TCL_OK)
		gdImageSetStyle(im, colors, ncolor);

 out:
	/* Free the colors. */
	if (colors != NULL)
		Tcl_Free((char *)colors);

#ifndef TCLOBJ
	if (colorArgv != &argv[4])
		Tcl_Free((char *)colorArgv);

#endif /* not TCLOBJ */
	return retval;
}

static int
#ifndef TCLOBJ
tclGdSetCmd(Tcl_Interp *interp, GdData *gdData, int argc, char *argv[])
#else /* TCLOBJ */
tclGdSetCmd(Tcl_Interp *interp, GdData *gdData, 
    int argc, Tcl_Obj *CONST objv[])
#endif /* TCLOBJ */
{
	gdImagePtr im;
	int color, x, y;
	
	/* Get the image pointer. */
#ifndef TCLOBJ
	im = *(gdImagePtr *)gdHandleXlate(interp, gdData->handleTbl, argv[2]);
#else /* TCLOBJ */
	im = *(gdImagePtr *)gdHandleXlate(interp, gdData->handleTbl, 
		Tcl_GetString(objv[2]));
#endif /* TCLOBJ */

	/* Get the color, x, y values. */
#ifndef TCLOBJ
	if (tclGd_GetColor(interp, argv[3], &color) != TCL_OK)
#else /* TCLOBJ */
	if (tclGd_GetColor(interp, objv[3], &color) != TCL_OK)
#endif /* TCLOBJ */
		return TCL_ERROR;
#ifndef TCLOBJ
	if (Tcl_GetInt(interp, argv[4], &x) != TCL_OK)
#else /* TCLOBJ */
	if (Tcl_GetIntFromObj(interp, objv[4], &x) != TCL_OK)
#endif /* TCLOBJ */
		return TCL_ERROR;
#ifndef TCLOBJ
	if (Tcl_GetInt(interp, argv[5], &y) != TCL_OK)
#else /* TCLOBJ */
	if (Tcl_GetIntFromObj(interp, objv[5], &y) != TCL_OK)
#endif /* TCLOBJ */
		return TCL_ERROR;

	/* Call the Set function. */
	gdImageSetPixel(im, x, y, color);

	return TCL_OK;
}

static int
#ifndef TCLOBJ
tclGdLineCmd(Tcl_Interp *interp, GdData *gdData, int argc, char *argv[])
#else /* TCLOBJ */
tclGdLineCmd(Tcl_Interp *interp, GdData *gdData, 
    int argc, Tcl_Obj *CONST objv[])
#endif /* TCLOBJ */
{
	gdImagePtr im;
	int color, x1, y1, x2, y2;
	
	/* Get the image pointer. */
#ifndef TCLOBJ
	im = *(gdImagePtr *)gdHandleXlate(interp, gdData->handleTbl, argv[2]);
#else /* TCLOBJ */
	im = *(gdImagePtr *)gdHandleXlate(interp, gdData->handleTbl,
		Tcl_GetString(objv[2]));
#endif /* TCLOBJ */

	/* Get the color, x, y values. */
#ifndef TCLOBJ
	if (tclGd_GetColor(interp, argv[3], &color) != TCL_OK)
#else /* TCLOBJ */
	if (tclGd_GetColor(interp, objv[3], &color) != TCL_OK)
#endif /* TCLOBJ */
		return TCL_ERROR;
#ifndef TCLOBJ
	if (Tcl_GetInt(interp, argv[4], &x1) != TCL_OK)
#else /* TCLOBJ */
	if (Tcl_GetIntFromObj(interp, objv[4], &x1) != TCL_OK)
#endif /* TCLOBJ */
		return TCL_ERROR;
#ifndef TCLOBJ
	if (Tcl_GetInt(interp, argv[5], &y1) != TCL_OK)
#else /* TCLOBJ */
	if (Tcl_GetIntFromObj(interp, objv[5], &y1) != TCL_OK)
#endif /* TCLOBJ */
		return TCL_ERROR;
#ifndef TCLOBJ
	if (Tcl_GetInt(interp, argv[6], &x2) != TCL_OK)
#else /* TCLOBJ */
	if (Tcl_GetIntFromObj(interp, objv[6], &x2) != TCL_OK)
#endif /* TCLOBJ */
		return TCL_ERROR;
#ifndef TCLOBJ
	if (Tcl_GetInt(interp, argv[7], &y2) != TCL_OK)
#else /* TCLOBJ */
	if (Tcl_GetIntFromObj(interp, objv[7], &y2) != TCL_OK)
#endif /* TCLOBJ */
		return TCL_ERROR;

	/* Call the appropriate Line function. */
	gdImageLine(im, x1, y1, x2, y2, color);

	return TCL_OK;
}

static int
#ifndef TCLOBJ
tclGdRectCmd(Tcl_Interp *interp, GdData *gdData, int argc, char *argv[])
#else /* TCLOBJ */
tclGdRectCmd(Tcl_Interp *interp, GdData *gdData, 
    int argc, Tcl_Obj *CONST objv[])
#endif /* TCLOBJ */
{
	gdImagePtr im;
	int color, x1, y1, x2, y2;
#ifdef TCLOBJ
	char *cmd;
#endif /* TCLOBJ */
	
	/* Get the image pointer. */
#ifndef TCLOBJ
	im = *(gdImagePtr *)gdHandleXlate(interp, gdData->handleTbl, argv[2]);
#else /* TCLOBJ */
	im = *(gdImagePtr *)gdHandleXlate(interp, gdData->handleTbl, 
		Tcl_GetString(objv[2]));
#endif /* TCLOBJ */

	/* Get the color, x, y values. */
#ifndef TCLOBJ
	if (tclGd_GetColor(interp, argv[3], &color) != TCL_OK)
#else /* TCLOBJ */
	if (tclGd_GetColor(interp, objv[3], &color) != TCL_OK)
#endif /* TCLOBJ */
		return TCL_ERROR;
#ifndef TCLOBJ
	if (Tcl_GetInt(interp, argv[4], &x1) != TCL_OK)
#else /* TCLOBJ */
	if (Tcl_GetIntFromObj(interp, objv[4], &x1) != TCL_OK)
#endif /* TCLOBJ */
		return TCL_ERROR;
#ifndef TCLOBJ
	if (Tcl_GetInt(interp, argv[5], &y1) != TCL_OK)
#else /* TCLOBJ */
	if (Tcl_GetIntFromObj(interp, objv[5], &y1) != TCL_OK)
#endif /* TCLOBJ */
		return TCL_ERROR;
#ifndef TCLOBJ
	if (Tcl_GetInt(interp, argv[6], &x2) != TCL_OK)
#else /* TCLOBJ */
	if (Tcl_GetIntFromObj(interp, objv[6], &x2) != TCL_OK)
#endif /* TCLOBJ */
		return TCL_ERROR;
#ifndef TCLOBJ
	if (Tcl_GetInt(interp, argv[7], &y2) != TCL_OK)
#else /* TCLOBJ */
	if (Tcl_GetIntFromObj(interp, objv[7], &y2) != TCL_OK)
#endif /* TCLOBJ */
		return TCL_ERROR;

	/* Call the appropriate rectangle function. */
#ifndef TCLOBJ
	if (argv[1][0] == 'r') 
#else /* TCLOBJ */
	cmd = Tcl_GetString(objv[1]);
	if (cmd[0] == 'r') 
#endif /* TCLOBJ */
		gdImageRectangle(im, x1, y1, x2, y2, color);
	else
		gdImageFilledRectangle(im, x1, y1, x2, y2, color);

	return TCL_OK;
}

static int
#ifndef TCLOBJ
tclGdArcCmd(Tcl_Interp *interp, GdData *gdData, int argc, char *argv[])
#else /* TCLOBJ */
tclGdArcCmd(Tcl_Interp *interp, GdData *gdData, 
    int argc, Tcl_Obj *CONST objv[])
#endif /* TCLOBJ */
{
	gdImagePtr im;
	int color, cx, cy, width, height, start, end;
#ifdef TCLOBJ
	char *cmd;
#endif /* TCLOBJ */
	
	/* Get the image pointer. */
#ifndef TCLOBJ
	im = *(gdImagePtr *)gdHandleXlate(interp, gdData->handleTbl, argv[2]);
#else /* TCLOBJ */
	im = *(gdImagePtr *)gdHandleXlate(interp, gdData->handleTbl,
		Tcl_GetString(objv[2]));
#endif /* TCLOBJ */

	/* Get the color, x, y values. */
#ifndef TCLOBJ
	if (tclGd_GetColor(interp, argv[3], &color) != TCL_OK)
#else /* TCLOBJ */
	if (tclGd_GetColor(interp, objv[3], &color) != TCL_OK)
#endif /* TCLOBJ */
		return TCL_ERROR;
#ifndef TCLOBJ
	if (Tcl_GetInt(interp, argv[4], &cx) != TCL_OK)
#else /* TCLOBJ */
	if (Tcl_GetIntFromObj(interp, objv[4], &cx) != TCL_OK)
#endif /* TCLOBJ */
		return TCL_ERROR;
#ifndef TCLOBJ
	if (Tcl_GetInt(interp, argv[5], &cy) != TCL_OK)
#else /* TCLOBJ */
	if (Tcl_GetIntFromObj(interp, objv[5], &cy) != TCL_OK)
#endif /* TCLOBJ */
		return TCL_ERROR;
#ifndef TCLOBJ
	if (Tcl_GetInt(interp, argv[6], &width) != TCL_OK)
#else /* TCLOBJ */
	if (Tcl_GetIntFromObj(interp, objv[6], &width) != TCL_OK)
#endif /* TCLOBJ */
		return TCL_ERROR;
#ifndef TCLOBJ
	if (Tcl_GetInt(interp, argv[7], &height) != TCL_OK)
#else /* TCLOBJ */
	if (Tcl_GetIntFromObj(interp, objv[7], &height) != TCL_OK)
#endif /* TCLOBJ */
		return TCL_ERROR;
#ifndef TCLOBJ
	if (Tcl_GetInt(interp, argv[8], &start) != TCL_OK)
#else /* TCLOBJ */
	if (Tcl_GetIntFromObj(interp, objv[8], &start) != TCL_OK)
#endif /* TCLOBJ */
		return TCL_ERROR;
#ifndef TCLOBJ
	if (Tcl_GetInt(interp, argv[9], &end) != TCL_OK)
#else /* TCLOBJ */
	if (Tcl_GetIntFromObj(interp, objv[9], &end) != TCL_OK)
#endif /* TCLOBJ */
		return TCL_ERROR;

	/* Call the appropriate arc function. */
#ifndef TCLOBJ
	if (argv[1][0] == 'a') 
#else /* TCLOBJ */
	cmd = Tcl_GetString(objv[1]);
	if (cmd[0] == 'a') 
#endif /* TCLOBJ */
		gdImageArc(im, cx, cy, width, height, start, end, color);
	else {
/*		gdImageFilledArc(im, cx, cy, width, height, start, end, color); */
#ifndef TCLOBJ
		sprintf(interp->result, "gdImageFilledArc not supported in gd1.2\n");
#else /* TCLOBJ */
		Tcl_SetResult(interp, "gdImageFilledArc not supported in gd1.2", TCL_STATIC);
#endif /* TCLOBJ */
		return TCL_ERROR;		  
	}

	return TCL_OK;
}

static int
#ifndef TCLOBJ
tclGdPolygonCmd(Tcl_Interp *interp, GdData *gdData, int argc, char *argv[])
#else /* TCLOBJ */
tclGdPolygonCmd(Tcl_Interp *interp, GdData *gdData, 
	int argc, Tcl_Obj *CONST objv[])
#endif /* TCLOBJ */
{
	gdImagePtr im;
	int color, npoints, i;
#ifndef TCLOBJ
	char **pointArgv = &argv[4];
#else /* TCLOBJ */
	Tcl_Obj **pointObjv = (Tcl_Obj **)(&objv[4]);
#endif /* TCLOBJ */
	gdPointPtr points = NULL;
	int retval = TCL_OK;
#ifdef TCLOBJ
	char *cmd;
#endif /* TCLOBJ */
	
	/* Get the image pointer. */
#ifndef TCLOBJ
	im = *(gdImagePtr *)gdHandleXlate(interp, gdData->handleTbl, argv[2]);
#else /* TCLOBJ */
	im = *(gdImagePtr *)gdHandleXlate(interp, gdData->handleTbl, 
		Tcl_GetString(objv[2]));
#endif /* TCLOBJ */

	/* Get the color, x, y values. */
#ifndef TCLOBJ
	if (tclGd_GetColor(interp, argv[3], &color) != TCL_OK)
#else /* TCLOBJ */
	if (tclGd_GetColor(interp, objv[3], &color) != TCL_OK)
#endif /* TCLOBJ */
		return TCL_ERROR;

	/* Figure out how many points in the list and allocate memory. */
	npoints = argc - 4;
	/* If only one argument, treat it as a list. */
	if (npoints == 1)
#ifndef TCLOBJ
		if (Tcl_SplitList(interp, argv[4], &npoints, &pointArgv) != TCL_OK)
#else /* TCLOBJ */
		if (Tcl_ListObjGetElements(interp, objv[4],
				&npoints, &pointObjv) != TCL_OK)
#endif /* TCLOBJ */
			return TCL_ERROR;

	/* Error check size of point list. */
	if (npoints % 2 != 0)
	{
#ifndef TCLOBJ
		sprintf(interp->result, "Number of coordinates must be even");
#else /* TCLOBJ */
		Tcl_SetResult(interp, "Number of coordinates must be even", TCL_STATIC);
#endif /* TCLOBJ */
		retval = TCL_ERROR;
		goto out;
	}

	/* Divide by 2 to get number of points, and final error check. */
	npoints /= 2;
	if (npoints < 3)
	{
#ifndef TCLOBJ
		sprintf(interp->result, "Must specify at least 3 points.");
#else /* TCLOBJ */
		Tcl_SetResult(interp, "Must specify at least 3 points.", TCL_STATIC);
#endif /* TCLOBJ */
		retval = TCL_ERROR;
		goto out;
	}

	points = (gdPointPtr)Tcl_Alloc(npoints * sizeof(gdPoint));
	if (points == NULL)
	{
#ifndef TCLOBJ
		sprintf(interp->result, "Memory allocation failed");
#else /* TCLOBJ */
		Tcl_SetResult(interp, "Memory allocation failed", TCL_STATIC);
#endif /* TCLOBJ */
		retval = TCL_ERROR;
		goto out;
	}

	/* Get the point values. */
	for (i = 0; i < npoints; i++)
#ifndef TCLOBJ
		if (Tcl_GetInt(interp, pointArgv[i*2], &points[i].x) != TCL_OK ||
			 Tcl_GetInt(interp, pointArgv[i*2+1], &points[i].y) != TCL_OK)
#else /* TCLOBJ */
		if (Tcl_GetIntFromObj(interp, pointObjv[i*2], &points[i].x) != TCL_OK ||
			 Tcl_GetIntFromObj(interp, pointObjv[i*2+1], &points[i].y) != TCL_OK)
#endif /* TCLOBJ */
		{
			retval = TCL_ERROR;
			break;
		}

	/* Call the appropriate polygon function. */
#ifndef TCLOBJ
	if (argv[1][0] == 'p') 
#else /* TCLOBJ */
	cmd = Tcl_GetString(objv[1]);
	if (cmd[0] == 'p') 
#endif /* TCLOBJ */
		gdImagePolygon(im, points, npoints, color);
	else
		gdImageFilledPolygon(im, points, npoints, color);

 out:
	/* Free the points. */
#ifndef TCLOBJ
	if (pointArgv != &argv[4])
		Tcl_Free((char *)pointArgv);
#endif /* not TCLOBJ */
	if (points != NULL)
		Tcl_Free((char *)points);

	return TCL_OK;
}

static int
#ifndef TCLOBJ
tclGdFillCmd(Tcl_Interp *interp, GdData *gdData, int argc, char *argv[])
#else /* TCLOBJ */
tclGdFillCmd(Tcl_Interp *interp, GdData *gdData,
	int argc, Tcl_Obj *CONST objv[])
#endif /* TCLOBJ */
{
	gdImagePtr im;
	int color, x, y, border;
	
	/* Get the image pointer. */
#ifndef TCLOBJ
	im = *(gdImagePtr *)gdHandleXlate(interp, gdData->handleTbl, argv[2]);
#else /* TCLOBJ */
	im = *(gdImagePtr *)gdHandleXlate(interp, gdData->handleTbl,
		Tcl_GetString(objv[2]));
#endif /* TCLOBJ */

	/* Get the color, x, y and possibly bordercolor values. */
#ifndef TCLOBJ
	if (tclGd_GetColor(interp, argv[3], &color) != TCL_OK)
#else /* TCLOBJ */
	if (tclGd_GetColor(interp, objv[3], &color) != TCL_OK)
#endif /* TCLOBJ */
		return TCL_ERROR;
#ifndef TCLOBJ
	if (Tcl_GetInt(interp, argv[4], &x) != TCL_OK)
#else /* TCLOBJ */
	if (Tcl_GetIntFromObj(interp, objv[4], &x) != TCL_OK)
#endif /* TCLOBJ */
		return TCL_ERROR;
#ifndef TCLOBJ
	if (Tcl_GetInt(interp, argv[5], &y) != TCL_OK)
#else /* TCLOBJ */
	if (Tcl_GetIntFromObj(interp, objv[5], &y) != TCL_OK)
#endif /* TCLOBJ */
		return TCL_ERROR;

	/* Call the appropriate fill function. */
	if (argc - 2 == 5) {
#ifndef TCLOBJ
		if (Tcl_GetInt(interp, argv[6], &border) != TCL_OK)
#else /* TCLOBJ */
		if (Tcl_GetIntFromObj(interp, objv[6], &border) != TCL_OK)
#endif /* TCLOBJ */
			return TCL_ERROR;
		gdImageFillToBorder(im, x, y, border, color);
	} else {
		gdImageFill(im, x, y, color);
	}

	return TCL_OK;
}

static int
#ifndef TCLOBJ
tclGdCopyCmd(Tcl_Interp *interp, GdData *gdData, int argc, char *argv[])
#else /* TCLOBJ */
tclGdCopyCmd(Tcl_Interp *interp, GdData *gdData, 
	int argc, Tcl_Obj *CONST objv[])
#endif /* TCLOBJ */
{
	gdImagePtr imdest, imsrc;
	int destx, desty, srcx, srcy, destw, desth, srcw, srch;
	
	/* Get the image pointer. */
#ifndef TCLOBJ
	imdest = *(gdImagePtr *)gdHandleXlate(interp, gdData->handleTbl, argv[2]);
	imsrc = *(gdImagePtr *)gdHandleXlate(interp, gdData->handleTbl, argv[3]);
#else /* TCLOBJ */
	imdest = *(gdImagePtr *)gdHandleXlate(interp, gdData->handleTbl, 
		Tcl_GetString(objv[2]));
	imsrc = *(gdImagePtr *)gdHandleXlate(interp, gdData->handleTbl,
		Tcl_GetString(objv[3]));
#endif /* TCLOBJ */

	/* Get the x, y, etc. values. */
#ifndef TCLOBJ
	if (Tcl_GetInt(interp, argv[4], &destx) != TCL_OK)
#else /* TCLOBJ */
	if (Tcl_GetIntFromObj(interp, objv[4], &destx) != TCL_OK)
#endif /* TCLOBJ */
		return TCL_ERROR;
#ifndef TCLOBJ
	if (Tcl_GetInt(interp, argv[5], &desty) != TCL_OK)
#else /* TCLOBJ */
	if (Tcl_GetIntFromObj(interp, objv[5], &desty) != TCL_OK)
#endif /* TCLOBJ */
		return TCL_ERROR;
#ifndef TCLOBJ
	if (Tcl_GetInt(interp, argv[6], &srcx) != TCL_OK)
#else /* TCLOBJ */
	if (Tcl_GetIntFromObj(interp, objv[6], &srcx) != TCL_OK)
#endif /* TCLOBJ */
		return TCL_ERROR;
#ifndef TCLOBJ
	if (Tcl_GetInt(interp, argv[7], &srcy) != TCL_OK)
#else /* TCLOBJ */
	if (Tcl_GetIntFromObj(interp, objv[7], &srcy) != TCL_OK)
#endif /* TCLOBJ */
		return TCL_ERROR;
#ifndef TCLOBJ
	if (Tcl_GetInt(interp, argv[8], &destw) != TCL_OK)
#else /* TCLOBJ */
	if (Tcl_GetIntFromObj(interp, objv[8], &destw) != TCL_OK)
#endif /* TCLOBJ */
		return TCL_ERROR;
#ifndef TCLOBJ
	if (Tcl_GetInt(interp, argv[9], &desth) != TCL_OK)
#else /* TCLOBJ */
	if (Tcl_GetIntFromObj(interp, objv[9], &desth) != TCL_OK)
#endif /* TCLOBJ */
		return TCL_ERROR;
	
	/* Call the appropriate copy function. */
	if (argc - 2 == 10) {
#ifndef TCLOBJ
		if (Tcl_GetInt(interp, argv[10], &srcw) != TCL_OK)
#else /* TCLOBJ */
		if (Tcl_GetIntFromObj(interp, objv[10], &srcw) != TCL_OK)
#endif /* TCLOBJ */
			return TCL_ERROR;
#ifndef TCLOBJ
		if (Tcl_GetInt(interp, argv[11], &srch) != TCL_OK)
#else /* TCLOBJ */
		if (Tcl_GetIntFromObj(interp, objv[11], &srch) != TCL_OK)
#endif /* TCLOBJ */
			return TCL_ERROR;

		gdImageCopyResized(imdest, imsrc, destx, desty, srcx, srcy,
							destw, desth, srcw, srch);
	}
	else
		gdImageCopy(imdest, imsrc, destx, desty, srcx, srcy, destw, desth);

	return TCL_OK;
}

static int
#ifndef TCLOBJ
tclGdGetCmd(Tcl_Interp *interp, GdData *gdData, int argc, char *argv[])
#else /* TCLOBJ */
tclGdGetCmd(Tcl_Interp *interp, GdData *gdData, 
	int argc, Tcl_Obj *CONST objv[])
#endif /* TCLOBJ */
{
	gdImagePtr im;
	int color, x, y;
	
	/* Get the image pointer. */
#ifndef TCLOBJ
	im = *(gdImagePtr *)gdHandleXlate(interp, gdData->handleTbl, argv[2]);
#else /* TCLOBJ */
	im = *(gdImagePtr *)gdHandleXlate(interp, gdData->handleTbl, 
		Tcl_GetString(objv[2]));
#endif /* TCLOBJ */

	/* Get the x, y values. */
#ifndef TCLOBJ
	if (Tcl_GetInt(interp, argv[3], &x) != TCL_OK)
#else /* TCLOBJ */
	if (Tcl_GetIntFromObj(interp, objv[3], &x) != TCL_OK)
#endif /* TCLOBJ */
		return TCL_ERROR;
#ifndef TCLOBJ
	if (Tcl_GetInt(interp, argv[4], &y) != TCL_OK)
#else /* TCLOBJ */
	if (Tcl_GetIntFromObj(interp, objv[4], &y) != TCL_OK)
#endif /* TCLOBJ */
		return TCL_ERROR;

	/* Call the Get function. */
	color = gdImageGetPixel(im, x, y);
#ifndef TCLOBJ
	sprintf(interp->result, "%d", color);
#else /* TCLOBJ */
	Tcl_SetObjResult(interp, Tcl_NewIntObj(color));
#endif /* TCLOBJ */
	return TCL_OK;
}

static int
#ifndef TCLOBJ
tclGdSizeCmd(Tcl_Interp *interp, GdData *gdData, int argc, char *argv[])
#else /* TCLOBJ */
tclGdSizeCmd(Tcl_Interp *interp, GdData *gdData, 
	int argc, Tcl_Obj *CONST objv[])
#endif /* TCLOBJ */
{
	gdImagePtr im;
#ifdef TCLOBJ
	char buf[30];
#endif /* TCLOBJ */

	/* Get the image pointer. */
#ifndef TCLOBJ
	im = *(gdImagePtr *)gdHandleXlate(interp, gdData->handleTbl, argv[2]);

	sprintf(interp->result, "%d %d", gdImageSX(im), gdImageSY(im));
#else /* TCLOBJ */
	im = *(gdImagePtr *)gdHandleXlate(interp, gdData->handleTbl, 
		Tcl_GetString(objv[2]));
#endif /* TCLOBJ */

#ifdef TCLOBJ
	sprintf(buf, "%d %d", gdImageSX(im), gdImageSY(im));
	Tcl_SetResult(interp, buf, TCL_VOLATILE);
#endif /* TCLOBJ */
	return TCL_OK;
}

static int
#ifndef TCLOBJ
tclGdTextCmd(Tcl_Interp *interp, GdData *gdData, int argc, char *argv[])
#else /* TCLOBJ */
tclGdTextCmd(Tcl_Interp *interp, GdData *gdData, 
	int argc, Tcl_Obj *CONST objv[])
#endif /* TCLOBJ */
{
	/* gd gdhandle color fontpathname size angle x y string */
	gdImagePtr			im;
	int					color, x, y;
	double				ptsize, angle;
	char				*error, buf[32], *font, *handle;
	int					i, brect[8], len;
	char                *str;
	/* unsigned char       *str; */

	/* Get the image pointer. (an invalid or null arg[2] will result in string
	size calculation but no rendering */
#ifndef TCLOBJ
	handle = argv[2];
#else /* TCLOBJ */
	handle = Tcl_GetString(objv[2]);
#endif /* TCLOBJ */
	if (! handle || *handle == '\0') {
		im = (gdImagePtr)NULL;
	} 
	else {
		im = *(gdImagePtr *)gdHandleXlate(interp, gdData->handleTbl, handle);
	}

	/* Get the color, values. */
#ifndef TCLOBJ
	if (tclGd_GetColor(interp, argv[3], &color) != TCL_OK) {
#else /* TCLOBJ */
	if (tclGd_GetColor(interp, objv[3], &color) != TCL_OK) {
#endif /* TCLOBJ */
		return TCL_ERROR;
	}

	/* Get point size */
#ifndef TCLOBJ
	if (Tcl_GetDouble(interp, argv[5], &ptsize) != TCL_OK) {
#else /* TCLOBJ */
	if (Tcl_GetDoubleFromObj(interp, objv[5], &ptsize) != TCL_OK) {
#endif /* TCLOBJ */
		return TCL_ERROR;
	}

	/* Get rotation (radians) */
#ifndef TCLOBJ
	if (Tcl_GetDouble(interp, argv[6], &angle) != TCL_OK) {
#else /* TCLOBJ */
	if (Tcl_GetDoubleFromObj(interp, objv[6], &angle) != TCL_OK) {
#endif /* TCLOBJ */
		return TCL_ERROR;
	}

	/* get x, y position */
#ifndef TCLOBJ
	if (Tcl_GetInt(interp, argv[7], &x) != TCL_OK) {
#else /* TCLOBJ */
	if (Tcl_GetIntFromObj(interp, objv[7], &x) != TCL_OK) {
#endif /* TCLOBJ */
		return TCL_ERROR;
	}
#ifndef TCLOBJ
	if (Tcl_GetInt(interp, argv[8], &y) != TCL_OK) {
#else /* TCLOBJ */
	if (Tcl_GetIntFromObj(interp, objv[8], &y) != TCL_OK) {
#endif /* TCLOBJ */
		return TCL_ERROR;
	}

#ifndef TCLOBJ
	str = argv[9];
	len = strlen(str);
	font = argv[4];
#else /* TCLOBJ */
	str = Tcl_GetStringFromObj(objv[9], &len);
	font = Tcl_GetString(objv[4]);
#endif /* TCLOBJ */

	error = gdttf(im, brect, color, font, ptsize, angle, x, y, str);

	if (error) {
#ifndef TCLOBJ
		Tcl_AppendResult(interp, error, NULL);
#else /* TCLOBJ */
		Tcl_SetResult(interp, error, TCL_VOLATILE);
#endif /* TCLOBJ */
		return TCL_ERROR;
	}
	for (i=0; i<8; i++) {
		sprintf(buf, "%d", brect[i]);
		Tcl_AppendElement(interp, buf);
	}
	return TCL_OK;
}

/* 
 * Initialize the package.
 */
int
Gdtclft_Init(Tcl_Interp *interp) 
{
	static GdData gdData;

#ifdef USE_TCL_STUBS
        if (Tcl_InitStubs(interp, TCL_VERSION, 0) == NULL) {
            return TCL_ERROR;
        }
#else
        if (Tcl_PkgRequire(interp, "Tcl", TCL_VERSION, 0) == NULL) {
            return TCL_ERROR;
        }
#endif
	/* if (Tcl_PkgProvide(interp, "Gdtclft", VERSION) != TCL_OK) { */
	if (Tcl_PkgProvide(interp, "Gdtclft", GDTCL_VERSION) != TCL_OK) {
		return TCL_ERROR;
	}

	GDHandleTable =
	gdData.handleTbl = gdHandleTblInit("gd", sizeof(gdImagePtr), 1);
    if (gdData.handleTbl == NULL)
    {
        Tcl_AppendResult(interp, "unable to create table for GD handles.", 0);
        return TCL_ERROR;
    }
		
#ifndef TCLOBJ
	Tcl_CreateCommand(interp, "gd", 
#else /* TCLOBJ */
	Tcl_CreateObjCommand(interp, "gd", 
#endif /* TCLOBJ */
		gdCmd, (ClientData)&gdData, (Tcl_CmdDeleteProc *)NULL);
	return TCL_OK;
}

int
Gdtclft_SafeInit(Tcl_Interp *interp) 
{
	return Gdtclft_Init(interp);
}
